/*
 * Kerio IP Filter library for userspace.
 * Copyright (C) Kerio Technologies s.r.o.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <pthread.h>
#include <stdarg.h>
#include <wchar.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <err.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/errno.h>

#include <kipf/api.h>

/* Globals */
static int fdesc = -1;

unsigned int kipf_get_sys_error(void)
{
	return errno;
}

kipf_status kipf_open(void)
{
	/*	Already opened ? */	
	if (fdesc != -1) {
		return KIPF_MAKE_STATUS(KIPF_ERROR_ALREADY_OPENED);
	}

	if ((fdesc = open(KIPF_IO_DEV, O_RDWR | O_CLOEXEC | O_NONBLOCK)) == -1) {
		return KIPF_MAKE_STATUS(KIPF_OPEN_DEVICE_ERROR);
	}

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

kipf_status kipf_close(unsigned int rule)
{
	if (fdesc == -1) {
		return KIPF_MAKE_STATUS(KIPF_NOT_OPENED);
	}	

	ioctl(fdesc, KIPF_IOCTL_SET_DEFAULT_TRAFFIC_RULE, &rule);

	close(fdesc);
	fdesc = -1;

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

kipf_status kipf_read(struct kipf_request *rreq)
{
	struct pollfd pfd;
	int ret = 0;

	if (fdesc == -1 || !rreq || rreq->count > MAX_RW_BUFFERS_COUNT) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

try_again:

	pfd.fd = fdesc;
	pfd.events = POLLIN | POLLERR;
	pfd.revents = 0;

	ret = poll(&pfd, 1, -1 /* infinite */);

	if (ret > 0) {

		if (pfd.revents & (POLLERR | POLLHUP | POLLNVAL)) {
			return KIPF_MAKE_STATUS(KIPF_READ_CANCELLED);			
		}

		if (pfd.revents & POLLIN) {

			if ((ret = read(fdesc, rreq, sizeof(struct kipf_request))) == -1) {	
				return KIPF_MAKE_STATUS(KIPF_READ_ERROR);
			}
	
			rreq->count = ret;

			return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
		}

	} else if (ret == 0) {
		/* timeout ? */
		return KIPF_MAKE_STATUS(KIPF_SELECT_ERROR);						
 	} else if (ret == -1) {
		if (errno == EINTR) {			
			goto try_again;
		}
		return KIPF_MAKE_STATUS(KIPF_SELECT_ERROR);				
	}			
	
	return KIPF_MAKE_STATUS(KIPF_SELECT_ERROR);		
}

kipf_status kipf_write(struct kipf_request *wreq)
{
	int ret = 0;

	if (fdesc == -1 || !wreq || wreq->count > MAX_RW_BUFFERS_COUNT) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if ((ret = write(fdesc, wreq, sizeof(struct kipf_request))) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_WRITE_ERROR);
	}

	wreq->count = ret;

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

kipf_status kipf_cancel_read(void)
{
	int ret = 0;

	if (fdesc == -1) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if ((ret = ioctl(fdesc, KIPF_IOCTL_CANCEL_PENDING_READ, NULL)) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}	

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);	
}

kipf_status kipf_wr_started(void)
{
	int ret = 0;

	if (fdesc == -1) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if ((ret = ioctl(fdesc, KIPF_IOCTL_WR_STARTED, NULL)) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}	

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);	
}

kipf_status kipf_get_driver_info(struct kipf_info *info)
{
	int ret = 0;

	if (fdesc == -1 || !info) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if ((ret = ioctl(fdesc, KIPF_IOCTL_GET_INFO, info)) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}	

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);		
}

kipf_status kipf_do_ioctl(unsigned int type, unsigned char *pkt, unsigned int pkt_len, struct kipf_route_info *route_info)
{
	IOCTL_REQ ioctlRequest;
	int ret = 0;

	if (fdesc == -1 || !pkt || ! route_info) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if (type != KIPF_ACTION_TX_IP_PACKET && type != KIPF_ACTION_RX_ETH_FRAME && type != KIPF_ACTION_TX_ETH_FRAME) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	ioctlRequest.action_type = type;

	ioctlRequest.pkt = pkt;
	ioctlRequest.pkt_len = pkt_len;
	ioctlRequest.route_info = *route_info;

	if ((ret = ioctl(fdesc, KIPF_IOCTL_HANDLE_PACKET, &ioctlRequest)) == -1) {
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

kipf_status kipf_handle_packet(struct kipf_packet *packet)
{
	IOCTL_REQ ioctlRequest;
	int ret = 0;

	if (fdesc == -1 || !packet) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	unsigned int type = packet->header.magic;

	if (type != KIPF_ACTION_ACCEPT && type != KIPF_ACTION_DROP && type != KIPF_ACTION_COMPLETE) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	ioctlRequest.action_type = type;
	ioctlRequest.packet = (KIPF_PTR) packet;

	if ((ret = ioctl(fdesc, KIPF_IOCTL_HANDLE_PACKET, &ioctlRequest)) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

kipf_status kipf_tx_ip_packet(unsigned char *pkt, unsigned int pkt_len, struct kipf_route_info *ri)
{
	return kipf_do_ioctl(KIPF_ACTION_TX_IP_PACKET, pkt, pkt_len, ri);
}

kipf_status kipf_rx_eth_frame(unsigned char *pkt, unsigned int pkt_len, unsigned int ifindex)
{
	struct kipf_route_info route_info = { {0}, ifindex};
	return kipf_do_ioctl(KIPF_ACTION_RX_ETH_FRAME, pkt, pkt_len, &route_info);
}

kipf_status kipf_tx_eth_frame(unsigned char *pkt, unsigned int pkt_len, unsigned int ifindex)
{
	struct kipf_route_info route_info = { {0}, ifindex};
	return kipf_do_ioctl(KIPF_ACTION_TX_ETH_FRAME, pkt, pkt_len, &route_info);
}

kipf_status kipf_set_nf_iface_mode(unsigned int ifindex, unsigned int enabled)
{
	NF_IFACE_MODE if_mode;
	int ret = 0;

	if (fdesc == -1) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if_mode.ifindex = ifindex;
	if_mode.mode = enabled ? NF_IFACE_MODE_ENABLED : NF_IFACE_MODE_DISABLED;

	if ((ret = ioctl(fdesc, KIPF_IOCTL_SET_NF_IFACE_MODE, &if_mode)) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

kipf_status kipf_gso_config(unsigned int enable)
{
	int ret = 0;

	if (fdesc == -1) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if ((ret = ioctl(fdesc, KIPF_IOCTL_GSO, &enable)) == -1) {
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}
